# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex utils module."""


import numpy as np
import numpy.typing as npt

from viskex.utils.scalar_type import ScalarType


def extract_part(
    values: npt.NDArray[ScalarType], name: str, part: str
) -> tuple[npt.NDArray[np.float64], str]:
    """Extract real or complex part from an array, and update the name to reflect this."""
    if np.issubdtype(ScalarType, np.complexfloating):
        if part == "real":
            values = values.real
            name = "real(" + name + ")"
        elif part == "imag":
            values = values.imag
            name = "imag(" + name + ")"
        else:
            raise RuntimeError(f"Invalid part {part}")
    else:
        if part != "real":
            raise RuntimeError(f"Invalid part {part}")
    return values, name



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex converter interfacing dolfinx."""

import typing

import dolfinx
import dolfinx.mesh
import dolfinx.plot
import numpy as np
import packaging.version
import pyvista
import ufl

from viskex.pyvista_converter import PyvistaConverter
from viskex.utils import extract_part

assert dolfinx.mesh.CellType.point not in dolfinx.plot._first_order_vtk
dolfinx.plot._first_order_vtk[dolfinx.mesh.CellType.point] = 1


class DolfinxConverter(PyvistaConverter[  # type: ignore[no-any-unimported]
    dolfinx.mesh.Mesh,
    typing.Union[dolfinx.fem.Function, tuple[ufl.core.expr.Expr, dolfinx.fem.FunctionSpace]]
]):
    """viskex converter interfacing dolfinx."""

    @classmethod
    def convert_mesh(cls, mesh: dolfinx.mesh.Mesh, dim: typing.Optional[int] = None) -> pyvista.UnstructuredGrid:
        """
        Convert a mesh stored in dolfinx.mesh.Mesh object.

        Parameters
        ----------
        mesh
            A dolfinx mesh to be converted.
        dim
            Convert entities associated to this dimension. If not provided, the topological dimension is used.

        Returns
        -------
        :
            A pyvista unstructured grid representing the mesh.
        """
        if dim is None:
            dim = mesh.topology.dim
        assert dim <= mesh.topology.dim

        # Convert the dolfinx mesh to a pyvista unstructured grid
        mesh.topology.create_connectivity(dim, dim)
        mesh.topology.create_connectivity(dim, mesh.topology.dim)
        num_cells = mesh.topology.index_map(dim).size_local + mesh.topology.index_map(dim).num_ghosts
        cell_entities = np.arange(num_cells, dtype=np.int32)
        pyvista_cells, cell_types, coordinates = dolfinx.plot.vtk_mesh(mesh, dim, cell_entities)
        return pyvista.UnstructuredGrid(pyvista_cells, cell_types, coordinates)

    @classmethod
    def convert_mesh_tags(
        cls, mesh: dolfinx.mesh.Mesh, mesh_tags: dolfinx.mesh.MeshTags, name: str
    ) -> pyvista.UnstructuredGrid:
        """
        Convert dolfinx.mesh.MeshTags.

        Parameters
        ----------
        mesh
            The dolfinx mesh associated to the provided mesh tags.
        mesh_tags
            A dolfinx mesh tags from which to extract mesh entities and their values.
        name
            Name to be assigned to the field containing the mesh entities values.

        Returns
        -------
        :
            A pyvista unstructured grid representing the mesh tags.
        """
        # Convert the dolfinx mesh to a pyvista unstructured grid, extracting entities of dimension mesh_tags.dim
        pyvista_grid = cls.convert_mesh(mesh, mesh_tags.dim)

        # Assign NaN values to missing indices
        assert np.invert(np.isnan(mesh_tags.values)).all(), "NaN is used as a placeholder for non-provided entities"
        values = np.full(pyvista_grid.n_cells, np.nan)
        for (index, value) in zip(mesh_tags.indices, mesh_tags.values):
            values[index] = value

        # Attach mesh tags to a cell field of the pyvista unstructured grid
        pyvista_grid.cell_data[name] = values
        pyvista_grid.set_active_scalars(name)
        return pyvista_grid

    @classmethod
    def convert_field(  # type: ignore[no-any-unimported]
        cls, field: typing.Union[
            dolfinx.fem.Function, tuple[ufl.core.expr.Expr, dolfinx.fem.FunctionSpace]
        ], name: str, part: str = "real"
    ) -> pyvista.UnstructuredGrid:
        """
        Convert a field stored in a dolfinx function, or a pair of UFL expression and dolfinx function space.

        Parameters
        ----------
        field
            Expression to be converted.
            If the expression is provided as a dolfinx function, such function will be converted.
            If the expression is provided as a tuple containing UFL expression and a dolfinx function space,
            the UFL expression will first be interpolated on the function space and then converted.
        name
            Name of the quantity stored in the field.
        part
            Part of the field (real or imag) to be converted. By default, the real part is converted.
            The argument is ignored when converting a real field.

        Returns
        -------
        :
            A pyvista unstructured grid representing the field.
        """
        # Interpolate the field if it is provided as an UFL expression
        if isinstance(field, tuple):
            expression, function_space = field
            interpolated_field = dolfinx.fem.Function(function_space)
            if packaging.version.Version(dolfinx.__version__) >= packaging.version.Version("0.10.0"):
                interpolation_points = function_space.element.interpolation_points
            else:
                interpolation_points = function_space.element.interpolation_points()  # type: ignore[operator, unused-ignore]
            interpolated_field.interpolate(dolfinx.fem.Expression(expression, interpolation_points))
        else:
            interpolated_field = field
            function_space = field.function_space

        # Get a pyvista unstructured grid based on the DOFs coordinates
        mesh = function_space.mesh
        dim = mesh.topology.dim
        mesh.topology.create_connectivity(dim, dim)
        num_cells = mesh.topology.index_map(dim).size_local + mesh.topology.index_map(dim).num_ghosts
        cell_entities = np.arange(num_cells, dtype=np.int32)
        pyvista_cells, cell_types, coordinates = dolfinx.plot.vtk_mesh(
            function_space, cell_entities)  # type: ignore[arg-type]
        pyvista_grid = pyvista.UnstructuredGrid(pyvista_cells, cell_types, coordinates)

        # Attach the field to the pyvista unstructured grid
        with interpolated_field.x.petsc_vec.localForm() as values:
            (values, name) = extract_part(values.array.copy(), name, part)
            if function_space.dofmap.index_map_bs > 1:
                values = values.reshape(-1, function_space.dofmap.index_map_bs)
                if values.shape[1] == 2:
                    values = np.insert(values, values.shape[1], 0.0, axis=1)
            pyvista_grid.point_data[name] = values
            if function_space.dofmap.index_map_bs == 1:
                pyvista_grid.set_active_scalars(name)
            else:
                pyvista_grid.set_active_vectors(name)
            return pyvista_grid



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex plotter interfacing dolfinx: user facing module, which automatically shows widgets."""

import typing

import dolfinx.fem
import dolfinx.mesh
import pyvista
import ufl

from viskex.dolfinx_plotter import DolfinxPlotter


def plot_mesh(
    mesh: dolfinx.mesh.Mesh, dim: typing.Optional[int] = None,
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a mesh stored in dolfinx.mesh.Mesh object.

    Parameters
    ----------
    mesh
        A dolfinx mesh to be plotted.
    dim
        Plot entities associated to this dimension. If not provided, the topological dimension is used.
    grid_filter
        A filter to be applied to the grid representing the mesh before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the mesh.
    """
    return DolfinxPlotter.plot_mesh(  # type: ignore[no-any-return]
        mesh, dim, grid_filter, **kwargs).show()  # type: ignore[no-untyped-call]


def plot_mesh_tags(
    mesh: dolfinx.mesh.Mesh, mesh_tags: dolfinx.mesh.MeshTags, name: str = "mesh tags",
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot dolfinx.mesh.MeshTags.

    Parameters
    ----------
    mesh
        The dolfinx mesh associated to the provided mesh tags.
    mesh_tags
        A dolfinx mesh tags from which to extract mesh entities and their values.
    name
        Name to be assigned to the field containing the mesh entities values.
    grid_filter
        A filter to be applied to the grid representing the mesh before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the mesh entities.
    """
    return DolfinxPlotter.plot_mesh_tags(  # type: ignore[no-any-return]
        mesh, mesh_tags, name, grid_filter, **kwargs).show()  # type: ignore[no-untyped-call]


def plot_scalar_field(  # type: ignore[no-any-unimported]
    scalar_field: typing.Union[
        dolfinx.fem.Function, tuple[ufl.core.expr.Expr, dolfinx.fem.FunctionSpace]
    ], name: str = "scalar", part: str = "real", warp_factor: float = 0.0,
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a scalar field stored in a dolfinx function, or a pair of UFL expression and dolfinx function space.

    Parameters
    ----------
    scalar_field
        Expression to be plotted, which contains a scalar field.
        If the expression is provided as a dolfinx function, such function will be plotted.
        If the expression is provided as a tuple containing UFL expression and a dolfinx function space,
        the UFL expression will first be interpolated on the function space and then plotted.
    name
        Name of the quantity stored in the scalar field.
    part
        Part of the field (real or imag) to be plotted. By default, the real part is plotted.
        The argument is ignored when plotting a real field.
    warp_factor
        If provided then the factor is used to produce a warped representation
        the field; if not provided then the scalar field will be plotted on the mesh.
    grid_filter
        A filter to be applied to the field representing the field before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the scalar field.
    """
    return DolfinxPlotter.plot_scalar_field(  # type: ignore[no-any-return]
        scalar_field, name, part, warp_factor, grid_filter, **kwargs).show()  # type: ignore[no-untyped-call]


def plot_vector_field(  # type: ignore[no-any-unimported]
    vector_field: typing.Union[
        dolfinx.fem.Function, tuple[ufl.core.expr.Expr, dolfinx.fem.FunctionSpace]
    ], name: str = "vector", part: str = "real", warp_factor: float = 0.0, glyph_factor: float = 0.0,
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a vector field stored in a dolfinx function, or a pair of UFL expression and dolfinx function space.

    Parameters
    ----------
    vector_field
        Expression to be plotted, which contains a vector field.
        If the expression is provided as a dolfinx function, such function will be plotted.
        If the expression is provided as a tuple containing UFL expression and a dolfinx function space,
        the UFL expression will first be interpolated on the function space and then plotted.
    name
        Name of the quantity stored in the vector field.
    part
        Part of the field (real or imag) to be plotted. By default, the real part is plotted.
        The argument is ignored when plotting a real field.
    warp_factor
        If provided then the factor is used to produce a warped representation of the field.
        If not provided then the magnitude of the vector field will be plotted on the mesh.
        The argument cannot be used if `glyph_factor` is also provided.
    glyph_factor
        If provided, the vector field is represented using a gylph, scaled by this factor.
        The argument cannot be used if `warp_factor` is also provided.
    grid_filter
        A filter to be applied to the field representing the field before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the vector field.
    """
    return DolfinxPlotter.plot_vector_field(  # type: ignore[no-any-return]
        vector_field, name, part, warp_factor, glyph_factor,  # type: ignore[no-untyped-call]
        grid_filter, **kwargs).show()



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex plotter interfacing firedrake: user facing module, which automatically shows widgets."""

import typing

import firedrake
import pyvista
import ufl

from viskex.firedrake_plotter import FiredrakePlotter


def plot_mesh(  # type: ignore[no-any-unimported]
    mesh: firedrake.MeshGeometry, dim: typing.Optional[int] = None,
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a mesh stored in firedrake.MeshGeometry object.

    Parameters
    ----------
    mesh
        A firedrake mesh to be plotted.
    dim
        Plot entities associated to this dimension. If not provided, the topological dimension is used.
    grid_filter
        A filter to be applied to the grid representing the mesh before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the mesh.
    """
    return FiredrakePlotter.plot_mesh(  # type: ignore[no-any-return]
        mesh, dim, grid_filter, **kwargs).show()  # type: ignore[no-untyped-call]


def plot_mesh_sets(  # type: ignore[no-any-unimported]
    mesh: firedrake.MeshGeometry, dim: int, name: str = "mesh sets",
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a cell set or a face sets of a given firedrake mesh.

    Parameters
    ----------
    mesh
        A firedrake mesh from which to extract mesh entities and their values.
    dim
        Extract entities associated to this dimension.
    name
        Name to be assigned to the field containing the mesh entities values.
    grid_filter
        A filter to be applied to the grid representing the mesh before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the mesh entities.
    """
    return FiredrakePlotter.plot_mesh_sets(  # type: ignore[no-any-return]
        mesh, dim, name, grid_filter, **kwargs).show()  # type: ignore[no-untyped-call]


def plot_scalar_field(  # type: ignore[no-any-unimported]
    scalar_field: typing.Union[
        firedrake.Function, tuple[ufl.core.expr.Expr, ufl.FunctionSpace]
    ], name: str = "scalar", part: str = "real", warp_factor: float = 0.0,
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a scalar field stored in a firedrake Function, or a pair of UFL Expression and firedrake FunctionSpace.

    Parameters
    ----------
    scalar_field
        Expression to be plotted, which contains a scalar field.
        If the expression is provided as a firedrake Function, such function will be plotted.
        If the expression is provided as a tuple containing UFL expression and a firedrake FunctionSpace,
        the UFL expression will first be interpolated on the function space and then plotted.
    name
        Name of the quantity stored in the scalar field.
    part
        Part of the field (real or imag) to be plotted. By default, the real part is plotted.
        The argument is ignored when plotting a real field.
    warp_factor
        If provided then the factor is used to produce a warped representation
        the field; if not provided then the scalar field will be plotted on the mesh.
    grid_filter
        A filter to be applied to the field representing the field before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the scalar field.
    """
    return FiredrakePlotter.plot_scalar_field(  # type: ignore[no-any-return]
        scalar_field, name, part, warp_factor, grid_filter, **kwargs).show()  # type: ignore[no-untyped-call]


def plot_vector_field(  # type: ignore[no-any-unimported]
    vector_field: typing.Union[
        firedrake.Function, tuple[ufl.core.expr.Expr, ufl.FunctionSpace]
    ], name: str = "vector", part: str = "real", warp_factor: float = 0.0, glyph_factor: float = 0.0,
    grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
    **kwargs: typing.Any  # noqa: ANN401
) -> pyvista.trame.jupyter.Widget:
    """
    Plot a vector field stored in a firedrake Function, or a pair of UFL Expression and firedrake FunctionSpace.

    Parameters
    ----------
    vector_field
        Expression to be plotted, which contains a vector field.
        If the expression is provided as a firedrake Function, such function will be plotted.
        If the expression is provided as a tuple containing UFL expression and a firedrake FunctionSpace,
        the UFL expression will first be interpolated on the function space and then plotted.
        Notice that the field will be interpolated to a P1 space before plotting.
    name
        Name of the quantity stored in the vector field.
    part
        Part of the field (real or imag) to be plotted. By default, the real part is plotted.
        The argument is ignored when plotting a real field.
    warp_factor
        If provided then the factor is used to produce a warped representation of the field.
        If not provided then the magnitude of the vector field will be plotted on the mesh.
        The argument cannot be used if `glyph_factor` is also provided.
    glyph_factor
        If provided, the vector field is represented using a gylph, scaled by this factor.
        The argument cannot be used if `warp_factor` is also provided.
    grid_filter
        A filter to be applied to the field representing the field before it is passed to pyvista.
        If not provided, no filter will be applied.
    kwargs
        Additional keyword arguments to be passed to pyvista.

    Returns
    -------
    :
        A pyvista widget representing a plot of the vector field.
    """
    return FiredrakePlotter.plot_vector_field(  # type: ignore[no-any-return]
        vector_field, name, part, warp_factor, glyph_factor,
        grid_filter, **kwargs).show()   # type: ignore[no-untyped-call]



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex converter interfacing firedrake."""

import typing

import firedrake
import firedrake.output
import numpy as np
import pyvista
import ufl

from viskex.pyvista_converter import PyvistaConverter
from viskex.utils import extract_part


class FiredrakeConverter(PyvistaConverter[  # type: ignore[no-any-unimported]
    firedrake.MeshGeometry,
    typing.Union[firedrake.Function, tuple[ufl.core.expr.Expr, ufl.FunctionSpace]]
]):
    """viskex converter interfacing firedrake."""

    # Conversion from UFL cell name and is_linear (True/False) to vtk cell type. Replicates the cells attribute
    # at the beginning of firedrake/output.py, with the exception that the boolean attribute is reversed
    # (this code uses is_linear, firedrake uses not is_linear)
    _ufl_cellname_to_vtk_celltype: typing.ClassVar[dict[tuple[str, bool], int]] = {
        ("point", True): 1,
        ("point", False): 1,
        ("interval", True): 3,
        ("interval", False): 68,
        ("triangle", True): 5,
        ("triangle", False): 69,
        ("quadrilateral", True): 9,
        ("quadrilateral", False): 70,
        ("tetrahedron", True): 10,
        ("tetrahedron", False): 71,
        ("hexahedron", True): 12,
        ("hexahedron", False): 72
    }

    # Conversion from UFL cell name and is_linear to vtk reordering lambda function.
    # The lambda function takes as input mesh element, and returns either a permutation, or None if
    # no permutation is required. Replicates part of the implementation in the function get_topology
    # in firedrake/output.py
    _ufl_cellname_to_vtk_permutation: typing.ClassVar[
        dict[tuple[str, bool], typing.Callable[[typing.Any], typing.Optional[list[int]]]]
    ] = {
        ("point", True): lambda ufl_element: None,
        ("point", False): lambda ufl_element: None,
        ("interval", True): lambda ufl_element: None,
        ("interval", False): firedrake.output.paraview_reordering.vtk_lagrange_interval_reorder,
        ("triangle", True): lambda ufl_element: None,
        ("triangle", False): firedrake.output.paraview_reordering.vtk_lagrange_triangle_reorder,
        ("quadrilateral", True): lambda ufl_element: [0, 2, 3, 1],
        ("quadrilateral", False): firedrake.output.paraview_reordering.vtk_lagrange_quad_reorder,
        ("tetrahedron", True): lambda ufl_element: None,
        ("tetrahedron", False): firedrake.output.paraview_reordering.vtk_lagrange_tet_reorder,
        ("hexahedron", True): lambda ufl_element: [0, 2, 3, 1, 4, 6, 7, 5],
        ("hexahedron", False): firedrake.output.paraview_reordering.vtk_lagrange_hex_reorder
    }

    # Conversion from UFL cell name of a mesh and topological dimension of an entity, to the UFL cell name
    # of the entity. Replicates the _sub_entity_celltypes attriute in ufl/cell.py.
    _tdim_cellname_to_dim_cellname: typing.ClassVar[dict[tuple[str, int], str]] = {
        ("point", 0): "point",
        ("interval", 1): "interval",
        ("interval", 0): "point",
        ("triangle", 2): "triangle",
        ("triangle", 1): "interval",
        ("triangle", 0): "point",
        ("quadrilateral", 2): "quadrilateral",
        ("quadrilateral", 1): "interval",
        ("quadrilateral", 0): "point",
        ("tetrahedron", 3): "tetrahedron",
        ("tetrahedron", 2): "triangle",
        ("tetrahedron", 1): "interval",
        ("tetrahedron", 0): "point",
        ("hexahedron", 3): "hexahedron",
        ("hexahedron", 2): "quadrilateral",
        ("hexahedron", 1): "interval",
        ("hexahedron", 0): "point"
    }

    @classmethod
    def convert_mesh(  # type: ignore[no-any-unimported]
        cls, mesh: firedrake.MeshGeometry, dim: typing.Optional[int] = None
    ) -> pyvista.UnstructuredGrid:
        """
        Convert a mesh stored in firedrake.MeshGeometry object.

        Parameters
        ----------
        mesh
            A firedrake mesh to be converted.
        dim
            Convert entities associated to this dimension. If not provided, the topological dimension is used.

        Returns
        -------
        :
            A pyvista unstructured grid representing the mesh.
        """
        tdim = mesh.topological_dimension()
        if dim is None:
            dim = tdim
        assert dim <= tdim

        # Determine connectivity from entites of dimension dim to vertices
        if dim == tdim:
            # The connectivity from cells to vertices is tabulated by firedrake
            connectivity = mesh.coordinates.cell_node_map().values_with_halo.copy()
        elif dim == 0 and tdim >= 2:
            # The connectivity from vertices to vertices is the identity, except for the 1D case
            # in which vertices are facets, and which will be handled in the next elif.
            vertices = mesh.coordinates.dat.data_ro_with_halos
            connectivity = np.arange(vertices.shape[0]).reshape(-1, 1)
        elif dim == tdim - 1:
            # Determine connectivity from facets to vertices: part of this code is inspired by
            # the triplot function in firedrake/plot.py
            topology = mesh.coordinates.function_space().finat_element.cell.get_topology()
            exterior_facet_local_ids = mesh.exterior_facets.local_facet_dat.data_ro_with_halos
            interior_facet_local_ids = mesh.interior_facets.local_facet_dat.data_ro_with_halos[:, :1].reshape(-1)
            facet_local_ids = np.concatenate((exterior_facet_local_ids, interior_facet_local_ids), dtype=np.int32)
            exterior_facet_node_map = mesh.coordinates.exterior_facet_node_map().values_with_halo
            interior_facet_node_map = mesh.coordinates.interior_facet_node_map()
            interior_facet_node_map = interior_facet_node_map.values_with_halo[:, :interior_facet_node_map.arity // 2]
            facet_node_map = np.concatenate((exterior_facet_node_map, interior_facet_node_map), dtype=np.int32)
            mask = np.zeros(facet_node_map.shape, dtype=bool)
            for mask_row, facet_local_id in enumerate(facet_local_ids):
                mask[mask_row, topology[tdim - 1][facet_local_id]] = True
            connectivity = facet_node_map[mask].reshape(-1, len(topology[tdim - 1][0]))
        else:
            raise NotImplementedError(f"The case dim={dim} and tdim={tdim} has not been implemented yet")

        # Permute connectivity information according to vtk ordering. The implementation is inspired by
        # the function get_topology in firedrake/output.py
        mesh_is_linear = firedrake.output.vtk_output.is_linear(mesh.coordinates.function_space())
        tdim_cellname = mesh.ufl_cell().cellname()
        dim_cellname = cls._tdim_cellname_to_dim_cellname[tdim_cellname, dim]
        permutation = cls._ufl_cellname_to_vtk_permutation[dim_cellname, mesh_is_linear](
            mesh.coordinates.function_space().ufl_element())
        if permutation is not None:
            connectivity = connectivity[:, permutation]

        # Convert the firedrake mesh to a pyvista unstructured grid
        connectivity = np.insert(connectivity, 0, values=connectivity.shape[1], axis=1)
        pyvista_types = np.full(connectivity.shape[0], cls._ufl_cellname_to_vtk_celltype[dim_cellname, mesh_is_linear])
        vertices = mesh.coordinates.dat.data_ro_with_halos
        (vertices, _) = extract_part(vertices, "vertices", "real")
        if tdim < 2:
            assert len(vertices.shape) == 1
            vertices = np.insert(vertices[:, np.newaxis], 1, values=0.0, axis=1)
        if tdim < 3:
            vertices = np.insert(vertices, 2, values=0.0, axis=1)
        return pyvista.UnstructuredGrid(connectivity.reshape(-1), pyvista_types, vertices)

    @classmethod
    def convert_mesh_sets(  # type: ignore[no-any-unimported]
        cls, mesh: firedrake.MeshGeometry, dim: int, name: str
    ) -> pyvista.UnstructuredGrid:
        """
        Convert cell sets or face sets of a given firedrake mesh.

        Parameters
        ----------
        mesh
            A firedrake mesh from which to extract mesh entities and their values.
        dim
            Extract entities associated to this dimension.
        name
            Name to be assigned to the field containing the mesh entities values.

        Returns
        -------
        :
            A pyvista unstructured grid representing the mesh sets.
        """
        tdim = mesh.topological_dimension()
        assert dim <= tdim

        # Convert the firedrake mesh to a pyvista unstructured grid, extracting entities of dimension dim
        pyvista_grid = cls.convert_mesh(mesh, dim)

        # Extract markers, assigning NaN to missing entities
        if dim == tdim:
            cells = mesh.coordinates.cell_node_map().values_with_halo
            assert cells.shape[0] == pyvista_grid.n_cells
            cell_markers = np.full(pyvista_grid.n_cells, np.nan)
            unique_cell_markers = mesh.topology_dm.getLabelIdIS(
                firedrake.cython.dmcommon.CELL_SETS_LABEL).indices.tolist()
            if len(unique_cell_markers) > 0:
                assert np.invert(np.isnan(unique_cell_markers)).all(), (
                    "NaN is used as a placeholder for unmarked cells")
                for cm in unique_cell_markers:
                    cell_markers[mesh.cell_subset(cm).indices] = cm
        elif dim == tdim - 1:
            facet_sizes = {
                facet_set_name: facet_set.measure_set(facet_set_name, "everywhere").total_size
                for (facet_set, facet_set_name) in zip(
                    (mesh.exterior_facets, mesh.interior_facets), ("exterior_facet", "interior_facet")
                )
            }
            assert sum(facet_sizes.values()) == pyvista_grid.n_cells
            facet_markers = np.full(pyvista_grid.n_cells, np.nan)
            unique_facet_markers = mesh.topology_dm.getLabelIdIS(
                firedrake.cython.dmcommon.FACE_SETS_LABEL).indices.tolist()
            if len(unique_facet_markers) > 0:
                assert np.invert(np.isnan(unique_facet_markers)).all(), (
                    "NaN is used as a placeholder for unmarked facets")
                for (facet_set, facet_set_name, offset) in zip(
                    (mesh.exterior_facets, mesh.interior_facets), ("exterior_facet", "interior_facet"),
                    (0, facet_sizes["exterior_facet"])
                ):
                    for fm in unique_facet_markers:
                        facet_indices_fm = offset + facet_set.measure_set(facet_set_name, fm).indices
                        facet_markers[facet_indices_fm] = fm
        else:
            raise NotImplementedError(f"The case dim={dim} and tdim={tdim} has not been implemented yet")

        # Attach mesh tags to a cell field of the pyvista unstructured grid
        if dim == tdim:
            pyvista_grid.cell_data[name] = cell_markers
        elif dim == tdim - 1:
            pyvista_grid.cell_data[name] = facet_markers
        pyvista_grid.set_active_scalars(name)
        return pyvista_grid

    @classmethod
    def convert_field(  # type: ignore[no-any-unimported]
        cls, field: typing.Union[
            firedrake.Function, tuple[ufl.core.expr.Expr, ufl.FunctionSpace]
        ], name: str, part: str = "real"
    ) -> pyvista.UnstructuredGrid:
        """
        Convert a field stored in a firedrake Function, or a pair of UFL Expression and firedrake FunctionSpace.

        Parameters
        ----------
        field
            Expression to be converted.
            If the expression is provided as a firedrake Function, such function will be converted.
            If the expression is provided as a tuple containing UFL expression and a firedrake FunctionSpace,
            the UFL expression will first be interpolated on the function space and then converted.
        name
            Name of the quantity stored in the field.
        part
            Part of the field (real or imag) to be converted. By default, the real part is converted.
            The argument is ignored when converting a real field.

        Returns
        -------
        :
            A pyvista unstructured grid representing the field.
        """
        # Interpolate the field if it is provided as an UFL expression
        if isinstance(field, tuple):
            expression, function_space = field
            interpolated_field = firedrake.Function(function_space)
            interpolated_field.interpolate(expression)
        else:
            interpolated_field = field
            function_space = field.function_space()

        # Get a firedrake mesh which has the same degree as the field. Note that the mesh attached to the function
        # space cannot be directly used, since it may be of a lower degree than the one associated to the field
        # (e.g. when the field is P^2, but the mesh is only P^1)
        assert function_space.rank in (0, 1)
        if function_space.rank == 0:
            field_vector_function_space = firedrake.VectorFunctionSpace(
                function_space.mesh(), function_space.ufl_element())
        else:
            field_vector_function_space = function_space
        mesh_coordinates = firedrake.Function(field_vector_function_space)
        mesh_coordinates.interpolate(function_space.mesh().coordinates)
        mesh = firedrake.Mesh(mesh_coordinates)

        # Convert the firedrake mesh to a pyvista unstructured grid
        pyvista_grid = cls.convert_mesh(mesh, mesh.topological_dimension())

        # Attach the field to the pyvista unstructured grid
        values = interpolated_field.dat.data_ro_with_halos
        (values, name) = extract_part(values, name, part)
        if function_space.rank == 1 and values.shape[1] == 2:
            values = np.insert(values, values.shape[1], 0.0, axis=1)
        pyvista_grid.point_data[name] = values
        if function_space.rank == 0:
            pyvista_grid.set_active_scalars(name)
        else:
            pyvista_grid.set_active_vectors(name)
        return pyvista_grid



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex converter interfacing pyvista."""

import pyvista

from viskex.base_converter import BaseConverter, Field, Mesh


class PyvistaConverter(BaseConverter[Mesh, Field, pyvista.UnstructuredGrid]):
    """viskex converter interfacing pyvista."""

    pass



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""Definition of default scalar type."""

import numpy as np

try:
    import petsc4py.PETSc
except ImportError:
    ScalarType = np.float64
else:
    ScalarType = petsc4py.PETSc.ScalarType  # type: ignore[attr-defined, misc, unused-ignore]



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex base plotter."""

import abc
import typing

Mesh = typing.TypeVar("Mesh")
ScalarFunction = typing.TypeVar("ScalarFunction")
VectorFunction = typing.TypeVar("VectorFunction")
OutputMesh = typing.TypeVar("OutputMesh")
OutputPlotter = typing.TypeVar("OutputPlotter")


class BasePlotter(abc.ABC, typing.Generic[Mesh, ScalarFunction, VectorFunction, OutputMesh, OutputPlotter]):
    """viskex base plotter."""

    @classmethod
    @abc.abstractmethod
    def plot_mesh(
        cls, mesh: Mesh, dim: typing.Optional[int] = None,
        grid_filter: typing.Optional[typing.Callable[[OutputMesh], OutputMesh]] = None,
        **kwargs: typing.Any  # noqa: ANN401
    ) -> OutputPlotter:
        """
        Plot a mesh.

        Parameters
        ----------
        mesh
            A mesh to be plotted.
        dim
            Plot entities associated to this dimension. If not provided, the topological dimension is used.
        grid_filter
            A filter to be applied to the grid representing the mesh before it is passed to the plotting backend.
            If not provided, no filter will be applied.
        kwargs
            Additional keyword arguments to be passed to the plotting backend.

        Returns
        -------
        :
            An output plotter representing the plot of the mesh.
        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def plot_scalar_field(
        cls, scalar_field: ScalarFunction, name: str = "scalar", part: str = "real", warp_factor: float = 0.0,
        grid_filter: typing.Optional[typing.Callable[[OutputMesh], OutputMesh]] = None,
        **kwargs: typing.Any  # noqa: ANN401
    ) -> OutputPlotter:
        """
        Plot a scalar field.

        Parameters
        ----------
        scalar_field
            A field to be plotted. The field must be a scalar field.
        name
            Name of the quantity stored in the scalar field.
        part
            Part of the field (real or imag) to be plotted. By default, the real part is plotted.
            The argument is ignored when plotting a real field.
        warp_factor
            If provided then the factor is used to produce a warped representation
            the field; if not provided then the scalar field will be plotted on the mesh.
        grid_filter
            A filter to be applied to the grid representing the field before it is pass to the plotting backend.
            If not provided, no filter will be applied.
        kwargs
            Additional keyword arguments to be passed to the plotting backend.

        Returns
        -------
        :
            A output plotter representing a plot of the scalar field.
        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def plot_vector_field(
        cls, vector_field: VectorFunction, name: str = "vector", part: str = "real", warp_factor: float = 0.0,
        glyph_factor: float = 0.0, grid_filter: typing.Optional[typing.Callable[[OutputMesh], OutputMesh]] = None,
        **kwargs: typing.Any  # noqa: ANN401
    ) -> OutputPlotter:
        """
        Plot a vector field stored in a dolfinx function, or a pair of UFL expression and dolfinx function space.

        Parameters
        ----------
        vector_field
            A field to be plotted. The field must be a vector field.
        name
            Name of the quantity stored in the vector field.
        part
            Part of the field (real or imag) to be plotted. By default, the real part is plotted.
            The argument is ignored when plotting a real field.
        warp_factor
            If provided then the factor is used to produce a warped representation of the field.
            If not provided then the magnitude of the vector field will be plotted on the mesh.
            The argument cannot be used if `glyph_factor` is also provided.
        glyph_factor
            If provided, the vector field is represented using a gylph, scaled by this factor.
            The argument cannot be used if `warp_factor` is also provided.
        grid_filter
            A filter to be applied to the grid representing the field before it is pass to the plotting backend.
            If not provided, no filter will be applied.
        kwargs
            Additional keyword arguments to be passed to the plotting backend.

        Returns
        -------
        :
            A output plotter representing a plot of the vector field.
        """
        pass  # pragma: no cover



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex base converter."""

import abc
import typing

Mesh = typing.TypeVar("Mesh")
Field = typing.TypeVar("Field")
OutputMesh = typing.TypeVar("OutputMesh")


class BaseConverter(abc.ABC, typing.Generic[Mesh, Field, OutputMesh]):
    """viskex base converter."""

    @classmethod
    @abc.abstractmethod
    def convert_mesh(cls, mesh: Mesh, dim: typing.Optional[int] = None) -> OutputMesh:
        """
        Convert a mesh.

        Parameters
        ----------
        mesh
            A mesh to be converted.
        dim
            Convert entities associated to this dimension. If not provided, the topological dimension is used.

        Returns
        -------
        :
            An unstructured grid representing the mesh.
        """
        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def convert_field(cls, field: Field, name: str, part: str = "real") -> OutputMesh:
        """
        Convert a field.

        Parameters
        ----------
        field
            Field to be converted.
        name
            Name of the quantity stored in the field.
        part
            Part of the field (real or imag) to be converted. By default, the real part is converted.
            The argument is ignored when converting a real field.

        Returns
        -------
        :
            An unstructured grid representing the field.
        """
        pass  # pragma: no cover



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""Limit cell data in the pyvista grid to specific values."""

import typing

import numpy as np
import numpy.typing as npt
import pyvista


def values_in(
    values: typing.Union[list[typing.Any], npt.NDArray[typing.Any]]
) -> typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]:
    """Limit cell data in the pyvista grid to specific values."""
    def _(grid: pyvista.UnstructuredGrid) -> pyvista.UnstructuredGrid:
        """Limit cell data in the pyvista grid to specific values: internal implementation."""
        name = grid.cell_data.active_scalars_name
        assert name is not None
        data = grid.cell_data.active_scalars
        assert data is not None
        not_equal_to_values, = np.where(~np.isin(data, values))
        data[not_equal_to_values] = np.nan  # type: ignore[index]
        grid.cell_data[name] = data
        return grid

    return _



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex utils module."""

from viskex.utils.extract_part import extract_part
from viskex.utils.scalar_type import ScalarType
from viskex.utils.values_in import values_in



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex main module."""

import importlib

from viskex.base_converter import BaseConverter
from viskex.base_plotter import BasePlotter
from viskex.pyvista_converter import PyvistaConverter
from viskex.pyvista_plotter import PyvistaPlotter

if importlib.util.find_spec("dolfinx"):
    import viskex.dolfinx
    from viskex.dolfinx_converter import DolfinxConverter
    from viskex.dolfinx_plotter import DolfinxPlotter

if importlib.util.find_spec("firedrake"):
    import viskex.firedrake
    from viskex.firedrake_converter import FiredrakeConverter
    from viskex.firedrake_plotter import FiredrakePlotter



# Copyright (C) 2023-2025 by the viskex authors
#
# This file is part of viskex.
#
# SPDX-License-Identifier: MIT
"""viskex plotter interfacing pyvista."""

import importlib.util
import os
import typing

import numpy as np
import pyvista

from viskex.base_plotter import BasePlotter

pyvista.set_plot_theme("document")  # type: ignore[no-untyped-call]
pyvista.global_theme.cmap = "jet"
pyvista.global_theme.color = "red"
pyvista.global_theme.edge_color = "black"
pyvista.global_theme.line_width = 2.0
pyvista.global_theme.nan_color = "lightgrey"
pyvista.global_theme.point_size = 10.0
pyvista.global_theme.show_edges = False
pyvista.global_theme.show_vertices = False

_jupyter_backend = "client"
_jupyter_notebook = None
if importlib.util.find_spec("google") and importlib.util.find_spec("google.colab"):  # pragma: no cover
    _jupyter_backend = "html"
    _jupyter_notebook = True
if "kaggle" in os.environ.get("KAGGLE_URL_BASE", ""):  # pragma: no cover
    _jupyter_backend = "html"
    _jupyter_notebook = True
_jupyter_backend = os.getenv("VISKEX_PYVISTA_BACKEND", _jupyter_backend)
assert _jupyter_backend in (
    "client", "html", "server", "trame",  # trame backends
    "static"  # static backend
)
pyvista.global_theme.jupyter_backend = _jupyter_backend
if _jupyter_notebook is not None:
    pyvista.global_theme.notebook = _jupyter_notebook
del _jupyter_backend
del _jupyter_notebook


class PyvistaPlotter(BasePlotter[
    tuple[pyvista.UnstructuredGrid, int], tuple[pyvista.UnstructuredGrid, int],
    tuple[pyvista.UnstructuredGrid, int], pyvista.UnstructuredGrid,
    pyvista.Plotter
]):
    """viskex plotter interfacing pyvista."""

    @classmethod
    def plot_mesh(
        cls, mesh: tuple[pyvista.UnstructuredGrid, int], dim: typing.Optional[int] = None,
        grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
        **kwargs: typing.Any  # noqa: ANN401
    ) -> pyvista.Plotter:
        """
        Plot a mesh stored in a pyvista.UnstructuredGrid.

        Parameters
        ----------
        mesh
            A pair containing the pyvista unstructured grid to be plotted and its topological dimension.
        dim
            Plot entities associated to this dimension. If not provided, the topological dimension is used.
        grid_filter
            A filter to be applied to the grid representing the mesh before it is passed to pyvista.Plotter.add_mesh.
            If not provided, no filter will be applied.
        kwargs
            Additional keyword arguments to be passed to pyvista.Plotter.add_mesh.

        Returns
        -------
        :
            A pyvista plotter representing a plot of the mesh.
        """
        grid, tdim = mesh
        if dim is None:
            dim = tdim

        # Filter the input grid
        if grid_filter is not None:
            grid = grid_filter(grid)

        # Create plotter
        plotter = pyvista.Plotter()  # type: ignore[no-untyped-call]

        # Terminate early if the grid is empty
        if not grid.n_points:
            return plotter

        # Determine whether cell data or point data have active scalars associated to the mesh
        active_scalars_name = {
            data_key: data.active_scalars_name
            for (data_key, data) in (("cell", grid.cell_data), ("point", grid.point_data))
        }

        # Reset to default cell color if scalar data have a single value
        if "cmap" not in kwargs:
            for data in (grid.cell_data, grid.point_data):
                if data.active_scalars is not None and not np.isnan(data.active_scalars).all():
                    if np.nanmin(data.active_scalars) == np.nanmax(data.active_scalars):
                        kwargs["cmap"] = [pyvista.global_theme.color.name]
                        break

        # Determine whether to show vertices: this happens by default in 1D, or when plotting points.
        show_vertices = kwargs.get("show_vertices", None)
        if show_vertices is None:
            if dim == 0:
                # No need to show vertices, since the grid itself is already composed by them
                show_vertices = False
            elif tdim < 2:
                show_vertices = True
            else:
                show_vertices = False
        if show_vertices:
            kwargs.pop("show_vertices", None)
            grid_points = pyvista.PolyData(grid.points)
            if dim == 0:
                if active_scalars_name["cell"] is not None:
                    # Do not override colors provided by cell data
                    vertex_color = None
                else:
                    vertex_color = kwargs.pop("color", pyvista.global_theme.color)
            else:
                if active_scalars_name["point"] is not None:
                    grid_points.point_data[active_scalars_name["point"]] = grid.point_data[
                        active_scalars_name["point"]]
                    # Do not override colors provided by point data
                    vertex_color = None
                else:
                    vertex_color = kwargs.pop("edge_color", pyvista.global_theme.edge_color)

        # Determine whether to show edges: this happens by default in 2D when plotting cells or in 3D
        # when plotting cells or faces.
        show_edges = kwargs.get("show_edges", None)
        if show_edges is None:
            if tdim > 1 and dim > 1:
                show_edges = True
            else:
                show_edges = False
        if show_edges:
            kwargs.pop("show_edges", None)
            edge_color = kwargs.pop("edge_color", pyvista.global_theme.edge_color)

        # Add grids to the plotter
        # Vertices and edges are manually added to plot, rather than using show_vertices and show_edges properties
        # because they lack support for high order meshes
        plotter.add_mesh(grid, **kwargs)
        if show_vertices:
            plotter.add_mesh(grid_points, color=vertex_color)
        if show_edges:
            plotter.add_mesh(grid.extract_all_edges(), color=edge_color)
        plotter.add_axes()  # type: ignore[call-arg]

        # Reset camera position in 1D and 2D
        if tdim < 3:
            plotter.camera_position = "xy"

        return plotter

    @classmethod
    def plot_scalar_field(
        cls, scalar_field: tuple[pyvista.UnstructuredGrid, int], name: str = "scalar", part: str = "real",
        warp_factor: float = 0.0,
        grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
        **kwargs: typing.Any  # noqa: ANN401
    ) -> pyvista.Plotter:
        """
        Plot a scalar field.

        Parameters
        ----------
        scalar_field
            A pair containing the pyvista unstructured grid to be plotted and its topological dimension.
            The grid must already have the scalar field to be plotted set as the active scalar.
        name
            This optional argument is never used.
        part
            This optional argument is never used.
        warp_factor
            If provided then the factor is used to produce a warped representation
            the field; if not provided then the scalar field will be plotted on the mesh.
        grid_filter
            A filter to be applied to the field representing the field before it is passed to pyvista.Plotter.add_mesh.
            If not provided, no filter will be applied.
        kwargs
            Additional keyword arguments to be passed to pyvista.Plotter.add_mesh.

        Returns
        -------
        :
            A pyvista plotter representing a plot of the scalar field.
        """
        grid, tdim = scalar_field

        # Filter the input grid
        if grid_filter is not None:
            grid = grid_filter(grid)

        # Warp the grid, if requested
        if warp_factor != 0.0:
            assert warp_factor > 0.0
            if tdim == 1:
                normal = [0, 1, 0]
            elif tdim == 2:
                normal = [0, 0, 1]
            else:
                normal = None
            warped_grid = grid.warp_by_scalar(factor=warp_factor, normal=normal)
            warped_tdim = tdim + 1
        else:
            warped_grid = grid
            warped_tdim = tdim

        # Call mesh plotter, with grid_filter=None because it was already applied in this function
        return cls.plot_mesh((warped_grid, warped_tdim), tdim, None, **kwargs)

    @classmethod
    def plot_vector_field(
        cls, vector_field: tuple[pyvista.UnstructuredGrid, int], name: str = "vector", part: str = "real",
        warp_factor: float = 0.0, glyph_factor: float = 0.0,
        grid_filter: typing.Optional[typing.Callable[[pyvista.UnstructuredGrid], pyvista.UnstructuredGrid]] = None,
        **kwargs: typing.Any  # noqa: ANN401
    ) -> pyvista.Plotter:
        """
        Plot a vector field.

        Parameters
        ----------
        vector_field
            A pair containing the pyvista unstructured grid to be plotted and its topological dimension.
            The grid must already have the vector field to be plotted set as the active vector.
        name
            This optional argument is never used.
        part
            This optional argument is never used.
        warp_factor
            If provided then the factor is used to produce a warped representation of the field.
            If not provided then the magnitude of the vector field will be plotted on the mesh.
            The argument cannot be used if `glyph_factor` is also provided.
        glyph_factor
            If provided, the vector field is represented using a gylph, scaled by this factor.
            The argument cannot be used if `warp_factor` is also provided.
        grid_filter
            A filter to be applied to the field representing the field before it is passed to pyvista.Plotter.add_mesh.
            If not provided, no filter will be applied.
        kwargs
            Additional keyword arguments to be passed to pyvista.Plotter.add_mesh.

        Returns
        -------
        :
            A pyvista plotter representing a plot of the vector field.
        """
        grid, tdim = vector_field

        # Filter the input grid
        if grid_filter is not None:
            grid = grid_filter(grid)

        # Warp the grid, if requested
        if warp_factor != 0.0:
            assert warp_factor > 0.0
            assert glyph_factor == 0.0
            warped_grid = grid.warp_by_vector(factor=warp_factor)
            warped_tdim = tdim
            warped_dim = tdim
        else:
            if glyph_factor != 0.0:
                # Show just mesh edges when adding glyphs
                warped_grid = grid.extract_all_edges()
                warped_dim = 1
            else:
                warped_grid = grid
                warped_dim = tdim
            warped_tdim = tdim

        # Call mesh plotter, with grid_filter=None because it was already applied in this function
        plotter = cls.plot_mesh((warped_grid, warped_tdim), warped_dim, None, **kwargs)

        # Add glyphs to the plot, if request
        if glyph_factor != 0.0:
            assert glyph_factor > 0.0
            assert warp_factor == 0.0
            glyphed_grid = grid.glyph(factor=glyph_factor)
            plotter.add_mesh(glyphed_grid)

        return plotter
